from __future__ import division #floating point division
import numpy as np
from numpy.random import rand, uniform, randint
from matplotlib import pylab as plt
from math import atan2, pi, sqrt, cos, sin, acos
from numpy.linalg import pinv, norm

n = 128
speed = 1.540 #mm/us speed of sound in tissue

RESOLUTION = speed/16
MAX_DELAY = 2

#uses least squares to perform pivot calibration given an array
#of rotation matrices and an array of translation vectors
def pivot_calibration( Rs, ps ):
    n = len( Rs )
    A = np.hstack( [np.vstack(Rs), np.vstack( [-np.eye(3)]*n )] )
    b = np.vstack(ps)
    return pinv(A).dot(b)
    
#make a rotation matrix with a given azimuth, elevation, and roll
def make_rotation(a, e, r):
    Z = np.array([[cos(a), -sin(a), 0],[sin(a), cos(a),0],[0,0,1]])
    Y = np.array([[cos(e), 0, sin(e)],[0,1,0],[-sin(e),0,cos(e)]])
    X = np.array([[1,0,0],[0,cos(r),-sin(r)],[0,sin(r),cos(r)]])
    return X.dot(Y).dot(Z);

#returns TODO FINISH EXPLAINING
def least_squares_generator( f_fun, J_fun, xs, ys):
    def least_squares( beta, shift_cut=1):
        r_i = [ y - f_fun(x,beta) for (x,y) in zip(xs,ys)]
        J = J_fun( xs, beta )
        d_beta = np.linalg.pinv(J).dot(r_i)
        return beta + shift_cut*d_beta, sum([r**2 for r in r_i])
    return least_squares

#calculates the Jacobian 
def myJacobian( x, beta ):
	J = np.zeros([len(x),3])
	for i in range(n):
		J[i,0] = 2*beta[0] - 2*beta[1]*x[i]*beta[2]
		J[i,1] = 2*beta[1]*x[i]**2 - 2*beta[0]*x[i]*beta[2]
		J[i,2] = -2*beta[0]*beta[1]*x[i]
	return J

def myJacobian_speed( x, beta ):
    m = len(beta)
    window = n//(m-3)
    J = np.zeros([len(x),len(beta)])
    for i in range(n):
        k = 3+i//window
        J[i,k] = 0.01*(beta[0]**2 - 2*x[i]*beta[0]*beta[1]*beta[2] + (beta[1]*x[i])**2)
        J[i,0] = (2*beta[0] - 2*beta[1]*x[i]*beta[2] )*beta[k]
        J[i,1] = (2*beta[1]*x[i]**2 - 2*beta[0]*x[i]*beta[2] )*beta[k]
        J[i,2] = (-2*beta[0]*beta[1]*x[i] )*beta[k]
    return J
    
def myf( x, beta ):
	return beta[0]**2 - 2*x*beta[0]*beta[1]*beta[2] + (beta[1]*x)**2

def myf_speed( x, beta ):
    window = n//(len(beta) - 3)
    return (beta[0]**2 - 2*x*beta[0]*beta[1]*beta[2] + (beta[1]*x)**2)*beta[3+x//window]
    
def calibrate( dist ):
    delta_i = 0.5
    L_i = 200
    theta_i = 90*pi/180
    beta = [L_i, delta_i, cos(theta_i*pi/180)]	
    least_squares = least_squares_generator( myf, myJacobian, range(n), [d**2 for d in dist] )
    for _ in range(40):
        beta, res = least_squares( beta )
    return beta
    
def calibrate_speed( time, window ):
    extra_parameters = n//window
    delta_i = 0.5
    L_i = 200
    theta_i = 90*pi/180
    beta = [L_i, delta_i, cos(theta_i*pi/180)]
    for i in range(extra_parameters):
        beta.append(1/speed**2)
    least_squares = least_squares_generator( myf_speed, myJacobian_speed, range(n), [t**2 for t in time] )
    for _ in range(20):
        #temp, res = beta, res = least_squares( beta )
        #beta[0:3] = temp[0:3]
        beta, res = least_squares( beta )
    return beta
    
def make_baseline( L_t, delta_t, theta_t ):
    pts = [ (delta_t*i*cos(theta_t), delta_t*i*sin(theta_t)) for i in range(n)]
    dist = [ sqrt( y**2 + (L_t-x)**2 ) for (x,y) in pts ]
    return dist 

def simple_test():
    sample_n = 20
    actual_values = np.zeros([sample_n,3])
    estimated_values = np.zeros([sample_n,3])
    error_record = np.zeros([sample_n,3])
    for i in range(sample_n):
        while( True ):
            L_t = uniform(150,400)
            delta_t = uniform(0.45,0.55)
            theta_t = uniform(0,180)*pi/180
            beta_t = np.array([L_t,delta_t,cos(theta_t)])
            
            dist = make_baseline( L_t, delta_t, theta_t )
            rands = randint(0,8,n)
            dist = [ round(d/RESOLUTION)*RESOLUTION+r*RESOLUTION for d,r in zip(dist,rands) ]
            beta = calibrate( dist )
            #do sanity checks
            if ( beta[2]<1 and beta[2]>-1 ):
                actual_values[i,:] = beta_t
                estimated_values[i,:] = beta
                error = (beta - beta_t)/beta_t*100
                error_record[i,:] = error
                break

    error_record = (estimated_values - actual_values)/actual_values * 100
    diff = estimated_values - actual_values
    plt.subplot(1,3,1)
    plt.boxplot( diff[:,0] )
    plt.title('Error in L (mm)')
    plt.subplot(1,3,2)
    plt.boxplot( diff[:,1] )
    plt.title('Error in delta (mm)')
    plt.subplot(1,3,3)
    plt.boxplot( diff[:,2])
    plt.title('Error in cos(theta)')
    plt.subplots_adjust(wspace=0.5)
    plt.show()

coef = 0.2/((127/2)*(127/2-127))
def speed_quadratic(x):
    return speed*(1 + coef*x*(x-127))

def speed_uniform(x):
    return speed
    
def complicated_test():
    sample_n = 20
    window = 128
    actual_values = np.zeros([sample_n,3])
    estimated_values = np.zeros([sample_n,3])
    error_record = np.zeros([sample_n,3])
    for i in range(sample_n):
        while( True ):
            L_t = uniform(150,400)
            delta_t = uniform(0.45,0.55)
            theta_t = uniform(0,180)*pi/180
            beta_t = np.array([L_t,delta_t,cos(theta_t)])
            
            dist = make_baseline( L_t, delta_t, theta_t )
            time = [ dist[x]/speed_uniform(x) for x in range(128)]
            rands = randint(0,8,n)
            time = [ round(t*16)/16+r/16 for t,r in zip(time,rands) ]
            beta = calibrate_speed( time, window )
            #do sanity checks
            if ( True or (beta[2]<1 and beta[2]>-1) ):
                actual_values[i,:] = beta_t[0:3]
                estimated_values[i,:] = beta[0:3]
                error = (beta[0:3] - beta_t[0:3])/beta_t[0:3]*100
                error_record[i,:] = error
                print beta, beta_t
                break
    error_record = (estimated_values - actual_values)/actual_values * 100
    diff = estimated_values - actual_values
    plt.plot(diff[:,2])
   # plt.plot(estimated_values[:,0])
   # plt.plot(actual_values[:,0])
 #   print beta
  # plt.boxplot( diff[:,1] )
   # plt.title('Distances error')
    plt.show()

def em_us_cal():
    offsets = []
    f_errs = []
    g_errs = []
    for _ in range(1000):
        m = 10
        f = uniform(-20,20,[3,1])
        v = uniform(-1,1,[3,1])
        g = f + v/sqrt( sum( v**2 ) )
        offset = uniform(-1,1,[3,1])
        offset = 0*offset/norm(offset)
        offsets.append( norm(offset) )
        
        Rs = []
        ps = []
        p_mag = []
        Ls = []
        angs = []
        for i in range(m):
            #make a random transformation
            a,e,r = uniform(0,2*pi,3)
            R = make_rotation(a,e,r)
            t = uniform(-300,300,[3,1])
            p = R.transpose().dot(t)
            wrong_p = R.transpose().dot(t-offset)
            Rs.append(R)
            ps.append(p)
            p_mag.append( sqrt( sum( wrong_p**2 ) ) )
            f_ = R.dot(f + p)
            g_ = R.dot(g + p)
            Ls.append( sqrt( sum ( f_**2 ) ) )
            angs.append( (f_ - g_).transpose().dot( f_ )/Ls[-1] )
        Ls = [l+e for l,e in zip(Ls,uniform(0.25,0.45,m))]
        angs = [a+e for a,e in zip(angs,uniform(-0.015,0.015,m))]
        b = []
        As = []
        for i in range(-1,m-1):
            b.append( Ls[i]**2 - p_mag[i]**2 - Ls[i+1]**2 + p_mag[i+1]**2 )
            b.append( Ls[i]*angs[i] - Ls[i+1]*angs[i+1])
            diff = ps[i] - ps[i+1]
            row1 = np.hstack([2*diff.transpose(),np.zeros([1,3])])
            row2 = np.hstack( [diff.transpose(), -diff.transpose()] )
            As.append( row1) 
            As.append( row2)
        A = np.vstack(As)
        b = np.array(b).reshape( [2*m,1] )
        x = pinv(A).dot(b)
        
        f_error = sum( sum( (x[0:3].transpose()-f)**2 ))
        g_error = sum( sum((x[3:].transpose() -g)**2 ))
        f_error = x[0:3] - f
        g_error = x[3:] - g
        f_errs.append( norm(f_error) )
        g_errs.append( norm(g_error) )
        c = x[0:3]-x[3:] - f + g

    #plt.plot(offsets,f_errs,'.')
    #plt.hold()
    #plt.xlabel('Hydrophone position error (mm)')
    #plt.ylabel('Error of f vector (mm)')
    #plt.title('EM-US calibration error dependence')
    #plt.axis('equal')
    
    plt.boxplot(f_errs,vert=False)
    plt.xlabel('Error of f vector (mm)')
    plt.title('EM-US calibration with no offset')
    plt.show()

def super_simple():
    pts = []
    a1 = rand(-200,200,[3,1])
    a2 = rand(-1,1,[3,1])
    a2 = a2/norm(a2)
    b1 = rand(-200,200,[3,1])
    b2 = rand(-1,1,[3,1])
    b2 = b2/norm(b2)
    
    for i in range(128):
        pts.append( a1+i*a2*0.5)
        pts.append( b1+i*b2*0.5)
    
    pts = rand(-200,200,[3,
#complicated_test()    
em_us_cal()
